import { FormControl, FormGroup, FormArray, FormBuilder } from '@angular/forms';
import { plainToClass, classToPlain, Expose, Transform } from 'class-transformer-global-storage';
import { validate, validateSync, Validate, ValidatorConstraint } from 'class-validator-multi-lang';
import cloneDeep from 'lodash.clonedeep';
import lodashGet from 'lodash.get';
import mergeWith from 'lodash.mergewith';
import lodashSet from 'lodash.set';
import { BehaviorSubject, of, ReplaySubject, from } from 'rxjs';
import { map, debounceTime, tap, delay, catchError } from 'rxjs/operators';
import { __decorate } from 'tslib';

const DEFAULT_CLASS_TRANSFORM_OPTIONS = {
    strategy: 'excludeAll',
};
const __dynamicControlOptions__ = '__dynamicControlOptions__';

/**
 * This function returns the global object across Node and browsers.
 *
 * Note: `globalThis` is the standardized approach however it has been added to
 * Node.js in version 12. We need to include this snippet until Node 12 EOL.
 */
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function getGlobal() {
    if (typeof globalThis !== 'undefined') {
        return globalThis;
    }
    if (typeof global !== 'undefined') {
        return global;
    }
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore: Cannot find name 'window'.
    if (typeof window !== 'undefined') {
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore: Cannot find name 'window'.
        return window;
    }
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore: Cannot find name 'self'.
    if (typeof self !== 'undefined') {
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore: Cannot find name 'self'.
        return self;
    }
    throw Error('Can not work with the global object');
}

const globalDynamicFormBuilderOptions = 'globalDynamicFormBuilderOptions';
function getGlobalDynamicFormBuilderOptionsSubject() {
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const global = getGlobal();
    if (!global[globalDynamicFormBuilderOptions]) {
        global[globalDynamicFormBuilderOptions] = new BehaviorSubject({});
    }
    return global[globalDynamicFormBuilderOptions];
}
function getGlobalDynamicFormBuilderOptions() {
    return (getGlobalDynamicFormBuilderOptionsSubject().value ||
        {});
}
function setGlobalDynamicFormBuilderOptions(dynamicFormBuilderOptions) {
    getGlobalDynamicFormBuilderOptionsSubject().next(dynamicFormBuilderOptions);
}

function setDynamicControlOptions(
// eslint-disable-next-line @typescript-eslint/no-explicit-any
object, options) {
    object[__dynamicControlOptions__] = options;
}
function removeDynamicControlOptions(
// eslint-disable-next-line @typescript-eslint/no-explicit-any
object) {
    delete object[__dynamicControlOptions__];
}
function getDynamicControlOptions(
// eslint-disable-next-line @typescript-eslint/no-explicit-any
object) {
    return object[__dynamicControlOptions__];
}
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function recursiveRemoveDynamicControlOptions(object) {
    if (!object || isPrimitiveClass(object) || isPrimitiveType(object)) {
        return object;
    }
    else {
        if (Array.isArray(object)) {
            return object.map((item) => recursiveRemoveDynamicControlOptions(item));
        }
        if (Object.getOwnPropertyDescriptor(object || {}, __dynamicControlOptions__)) {
            delete object[__dynamicControlOptions__];
        }
        Object.keys(object).forEach((key) => {
            if (Object.getOwnPropertyDescriptor(object[key] || {}, __dynamicControlOptions__)) {
                delete object[key][__dynamicControlOptions__];
            }
        });
        return object;
    }
}
function replaceLastPropertyToDynamicControlOptionsByDotPath(controlPath, 
// eslint-disable-next-line @typescript-eslint/no-explicit-any
object, 
// eslint-disable-next-line @typescript-eslint/no-explicit-any
rootValue, control, metadata) {
    const newControlPath = [
        ...controlPath
            .split('.')
            .filter((v, i) => i < controlPath.split('.').length - 1),
        __dynamicControlOptions__,
    ].join('.');
    lodashSet(
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    object, newControlPath, {
        rootValue,
        controlPath,
        control,
        metadata,
    });
}
function setCustomDataToRootFormGroup(
// eslint-disable-next-line @typescript-eslint/no-explicit-any
control, key, value) {
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    control.root[key] = value;
}
function getCustomDataToRootFormGroup(
// eslint-disable-next-line @typescript-eslint/no-explicit-any
control, key, defaultValue) {
    if (Object.getOwnPropertyDescriptor(control?.root || {}, key)) {
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        return control?.root[key];
    }
    return defaultValue;
}
function transformValidationErrorsToClassValidatorErrors(errors) {
    const customErrors = {};
    errors.forEach((error) => {
        if (error && error.constraints !== undefined) {
            Object.keys(error.constraints).forEach((key) => {
                if (!customErrors[error.property]) {
                    customErrors[error.property] = {};
                }
                if (!customErrors[error.property].messages) {
                    customErrors[error.property].messages = [];
                }
                if (error.constraints &&
                    customErrors[error.property].messages?.indexOf(error.constraints[key]) === -1) {
                    customErrors[error.property].messages?.push(error.constraints[key]);
                }
            });
        }
        if (error.children !== undefined && error.children.length) {
            if (!customErrors[error.property]) {
                customErrors[error.property] = {};
            }
            if (!customErrors[error.property].children) {
                customErrors[error.property].children = {};
            }
            customErrors[error.property].children =
                transformValidationErrorsToClassValidatorErrors(error.children);
        }
    });
    return customErrors;
}
function transformClassValidatorErrorsToShortValidationErrors(errors) {
    const customErrors = {};
    Object.keys(errors).forEach((key) => {
        if (errors[key]?.messages?.length) {
            if (!customErrors[key]) {
                customErrors[key] = {};
            }
            customErrors[key] = errors[key].messages;
        }
        if (Object.keys(errors[key]?.children || {}).length) {
            if (!customErrors[key]) {
                customErrors[key] = {};
            }
            customErrors[key] = {
                ...transformClassValidatorErrorsToShortValidationErrors(errors[key].children || {}),
            };
        }
    });
    return customErrors;
}
function mergeErrors(externalErrors, validationErrors) {
    const clonedExternalErrors = cloneDeep(externalErrors);
    const clonedValidationErrors = cloneDeep(validationErrors);
    return mergeWith(clonedExternalErrors, clonedValidationErrors, (objValue, srcValue) => {
        if (Array.isArray(objValue) && !Array.isArray(srcValue)) {
            return [...objValue, ...Object.keys(srcValue || {})].filter(Boolean);
        }
        if (Array.isArray(srcValue) && !Array.isArray(objValue)) {
            return [...Object.keys(objValue || {}), ...srcValue].filter(Boolean);
        }
        if (canMerge()) {
            return objValue.concat(srcValue);
        }
        return undefined;
        function canMerge() {
            return (Array.isArray(objValue) &&
                Array.isArray(srcValue) &&
                !objValue.find((objItem) => srcValue.includes(objItem)));
        }
    });
}
function collectDynamicFormGroupErrors(control) {
    if (control.controls) {
        return {
            ...Object.keys(control.controls).reduce((all, key) => {
                const childErrors = collectDynamicFormGroupErrors(control.controls[key]);
                if (childErrors && Object.keys(childErrors).length > 0) {
                    all = {
                        ...all,
                        [key]: {
                            ...childErrors,
                        },
                    };
                }
                return all;
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
            }, {}),
        };
    }
    else {
        // todo: need check
        const key = control?.classTransformMetadata?.propertyName;
        const childErrors = control.errors;
        if (key &&
            childErrors?.[key]?.messages &&
            Array.isArray(childErrors[key]?.messages)) {
            delete childErrors[key];
        }
        return childErrors;
    }
}
function isPrimitiveClass(classType) {
    return (classType &&
        (classType === Number ||
            classType === String ||
            classType === Boolean ||
            classType === Date));
}
function isPrimitiveType(object) {
    return (object &&
        (typeof object === 'number' ||
            typeof object === 'string' ||
            typeof object === 'boolean'));
}

function createFormControls({ classType, formBuilder, form, rootFormGroup, metadata, defaultValue, dynamicFormBuilderOptions, }) {
    const isRoot = !rootFormGroup;
    let dynamicForm = form;
    if (!dynamicForm) {
        dynamicForm = formBuilder.group({});
    }
    if (isRoot) {
        rootFormGroup = dynamicForm;
        setupDynamicFormBuilderOptions({
            dynamicForm,
            dynamicFormBuilderOptions,
        });
    }
    if (dynamicFormBuilderOptions || rootFormGroup?.dynamicFormBuilderOptions) {
        setupClassTransformMetadata({
            classType,
            dynamicForm,
            defaultMetadata: metadata,
            dynamicFormBuilderOptions: 
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            dynamicFormBuilderOptions || rootFormGroup.dynamicFormBuilderOptions,
        });
    }
    createAllFormGroupChildrenControls({
        dynamicForm,
        defaultValue,
        formBuilder,
        rootFormGroup,
        dynamicFormBuilderOptions,
    });
    addDynamicMethodsToFormGroup(dynamicForm);
    if (isRoot) {
        subscribeToRootFormGroupValueChanges(dynamicForm);
        subscribeToGlobalDynamicFormBuilderOptionsChange(formBuilder, dynamicForm);
        return addDynamicMethodsToRootFormGroup({
            dynamicForm,
            rootFormGroup,
            formBuilder,
        });
    }
    else {
        return dynamicForm;
    }
}
function setValuesForControls(formBuilder, form, value) {
    const metadata = form.classTransformMetadata;
    if (!metadata) {
        throw new Error(`classTransformMetadata not found in form group`);
    }
    if (metadata.classType &&
        !isPrimitiveClass(metadata.classType) &&
        !(value instanceof metadata.classType)) {
        // need for multi types
        if (metadata.withAncestors &&
            metadata.parent?.classType &&
            metadata.propertyName) {
            const isArray = metadata.parent.properties.find((prop) => prop.propertyName === metadata.propertyName)?.isArray || false;
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            const parentClassType = metadata.parent.classType;
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            const parentObject = {
                [metadata.propertyName]: isArray ? [value] : value,
            };
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            const parentJson = plainToClass(parentClassType, parentObject, 
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            form.root.dynamicFormBuilderOptions
                .classTransformOptions);
            value = isArray
                ? parentJson[metadata.propertyName][0]
                : parentJson[metadata.propertyName];
        }
        else {
            value = plainToClass(metadata.classType, value, 
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            form.root.dynamicFormBuilderOptions
                .classTransformOptions);
        }
    }
    metadata.properties.forEach((metadataItem) => {
        if (!metadataItem.ignore) {
            if (metadataItem.propertyName) {
                if (!metadataItem.ignore) {
                    if (!metadataItem.isArray) {
                        const control = form.controls[metadataItem.propertyName];
                        if (!metadataItem.classType ||
                            isPrimitiveClass(metadataItem.classType)) {
                            try {
                                control.patchValue(value !== undefined &&
                                    Object.getOwnPropertyDescriptor(value || {}, metadataItem.propertyName)
                                    ? value[metadataItem.propertyName]
                                    : null, { emitEvent: false });
                            }
                            catch (err) {
                                console.log({ err, metadataItem, form });
                            }
                        }
                        else {
                            try {
                                setValuesForControls(formBuilder, 
                                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                                form.controls[metadataItem.propertyName], value !== undefined
                                    ? value[metadataItem.propertyName]
                                    : {});
                            }
                            catch (err) {
                                console.log({ err, metadataItem, form });
                                throw err;
                            }
                        }
                    }
                    else {
                        const arrayControl = form.controls[metadataItem.propertyName];
                        if (arrayControl?.classTransformMetadata?.isArray &&
                            arrayControl?.controls.length > 0) {
                            let formArrayLength = arrayControl.controls.length;
                            while (formArrayLength !== 0) {
                                arrayControl.removeAt(0);
                                formArrayLength--;
                            }
                        }
                        if (!metadataItem.classType ||
                            isPrimitiveClass(metadataItem.classType)) {
                            try {
                                ((Object.getOwnPropertyDescriptor(value || {}, metadataItem.propertyName)
                                    ? value[metadataItem.propertyName]
                                    : null) || []).forEach((item) => {
                                    if (metadataItem.propertyName) {
                                        arrayControl.push(formBuilder.control(item));
                                    }
                                });
                            }
                            catch (err) {
                                console.log({ err, metadataItem, form });
                            }
                        }
                        else {
                            try {
                                ((Object.getOwnPropertyDescriptor(value || {}, metadataItem.propertyName)
                                    ? value[metadataItem.propertyName]
                                    : null) || [])
                                    // eslint-disable-next-line @typescript-eslint/no-explicit-any
                                    .forEach((item) => {
                                    // need for multi types
                                    const itemKeys = Object.keys(item);
                                    if (metadataItem.withAncestors) {
                                        metadataItem.properties.forEach((prop) => {
                                            if (prop.propertyName &&
                                                !itemKeys.includes(prop.propertyName)) {
                                                item[prop.propertyName] = null;
                                            }
                                        });
                                    }
                                    if (metadataItem.propertyName) {
                                        const control = createFormControls({
                                            classType: metadataItem.classType,
                                            formBuilder,
                                            metadata: metadataItem,
                                            defaultValue: item,
                                            rootFormGroup: form.root,
                                            dynamicFormBuilderOptions: form.root.dynamicFormBuilderOptions,
                                        });
                                        arrayControl.push(control);
                                    }
                                });
                            }
                            catch (err) {
                                console.log({ err, metadataItem, form });
                            }
                        }
                    }
                }
            }
        }
    });
}
function setValidatorsToControls(formBuilder, form, rootFormGroup) {
    if (!rootFormGroup) {
        rootFormGroup = form.root;
    }
    addCommonAsyncValidatorToRootForm(rootFormGroup);
    const metadata = form.classTransformMetadata;
    if (!metadata) {
        throw new Error('classTransformMetadata not found in form group');
    }
    metadata.properties.forEach((metadataItem) => {
        if (!metadataItem.ignore && metadataItem.propertyName) {
            if (!metadataItem.isArray) {
                if (!metadataItem.classType ||
                    isPrimitiveClass(metadataItem.classType)) {
                    if (!rootFormGroup?.commonAsyncValidator) {
                        throw new Error('commonAsyncValidator not set');
                    }
                    try {
                        if (form.controls[metadataItem.propertyName]) {
                            form.controls[metadataItem.propertyName].classTransformMetadata = metadataItem;
                            if (!form.controls[metadataItem.propertyName].hasAsyncValidator(rootFormGroup.commonAsyncValidator)) {
                                form.controls[metadataItem.propertyName].setAsyncValidators(rootFormGroup.commonAsyncValidator); /*
                                form.controls[
                                  metadataItem.propertyName
                                ].updateValueAndValidity();*/
                            }
                            if (rootFormGroup.dynamicFormBuilderOptions.angularValidators) {
                                const { angularFormValidatorPath } = findControlPaths(form.controls[metadataItem.propertyName]);
                                const angularValidators = lodashGet(rootFormGroup.dynamicFormBuilderOptions.angularValidators, angularFormValidatorPath
                                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                                );
                                if (angularValidators) {
                                    let asyncValidators = [];
                                    if (angularValidators.asyncValidators) {
                                        asyncValidators = angularValidators.asyncValidators;
                                        if (typeof asyncValidators === 'function') {
                                            asyncValidators = [asyncValidators];
                                        }
                                    }
                                    let validators = angularValidators?.validators || angularValidators;
                                    if (typeof validators === 'function') {
                                        validators = [validators];
                                    }
                                    form.controls[metadataItem.propertyName].setAsyncValidators([
                                        ...asyncValidators,
                                        ...validators
                                            .filter(Boolean)
                                            .map((validator) => (control) => Promise.resolve(validator(control))),
                                        rootFormGroup.commonAsyncValidator,
                                    ]); /*
                                    form.controls[
                                      metadataItem.propertyName
                                    ].updateValueAndValidity();*/
                                }
                            }
                        }
                    }
                    catch (err) {
                        console.log({ err, metadataItem, form });
                    }
                }
                else {
                    try {
                        setValidatorsToControls(formBuilder, form.controls[metadataItem.propertyName], rootFormGroup);
                    }
                    catch (err) {
                        console.log({ err, metadataItem, form });
                    }
                }
            }
            else {
                if (!metadataItem.classType ||
                    isPrimitiveClass(metadataItem.classType)) {
                    if (!rootFormGroup?.commonAsyncValidator) {
                        throw new Error('commonAsyncValidator not set');
                    }
                    try {
                        const controls = form.controls[metadataItem.propertyName].controls;
                        (!Array.isArray(controls) && metadataItem.isArray
                            ? []
                            : controls).forEach((control, propertyIndex) => {
                            control.classTransformMetadata = {
                                ...metadataItem,
                                propertyIndex,
                                isArray: false,
                            };
                            if (!rootFormGroup) {
                                throw Error('rootFormGroup not set');
                            }
                            if (!control.hasAsyncValidator(rootFormGroup.commonAsyncValidator)) {
                                control.setAsyncValidators(rootFormGroup.commonAsyncValidator);
                                /* control.updateValueAndValidity(); */
                            }
                        });
                        if (!rootFormGroup) {
                            throw Error('rootFormGroup not set');
                        }
                        if (!form.controls[metadataItem.propertyName].hasAsyncValidator(rootFormGroup.commonAsyncValidator)) {
                            form.controls[metadataItem.propertyName].setAsyncValidators(rootFormGroup.commonAsyncValidator);
                            /* (
                              form.controls[metadataItem.propertyName] as DynamicFormArray
                            ).updateValueAndValidity(); */
                        }
                    }
                    catch (err) {
                        console.log({ err, metadataItem, form });
                    }
                }
                else {
                    try {
                        const controls = form.controls[metadataItem.propertyName].controls;
                        (!Array.isArray(controls) && metadataItem.isArray
                            ? []
                            : controls).forEach((control, propertyIndex) => {
                            control.classTransformMetadata = {
                                ...metadataItem,
                                propertyIndex,
                                isArray: false,
                            };
                            const controlValue = control.value || {};
                            // needed for extend controls if it does not exists, but the value has property with that name
                            if (metadataItem.classType &&
                                !isPrimitiveType(metadataItem.classType) &&
                                metadataItem.properties.length >
                                    Object.keys(controlValue).length) {
                                metadataItem.properties.forEach((prop) => {
                                    if (prop.propertyName &&
                                        !Object.getOwnPropertyDescriptor(controlValue, prop.propertyName)) {
                                        if (prop.classType && !isPrimitiveType(prop.classType)) {
                                            controlValue[prop.propertyName] = {};
                                            if (
                                            // eslint-disable-next-line @typescript-eslint/no-explicit-any
                                            !control.controls[prop.propertyName]) {
                                                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                                                control.addControl(prop.propertyName, formBuilder.group(controlValue[prop.propertyName]));
                                            }
                                        }
                                        else {
                                            if (prop.isArray) {
                                                controlValue[prop.propertyName] = [];
                                                if (
                                                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                                                !control.controls[prop.propertyName]) {
                                                    // eslint-disable-next-line @typescript-eslint/no-explicit-any
                                                    control.addControl(prop.propertyName, formBuilder.array(controlValue[prop.propertyName]));
                                                }
                                            }
                                            else {
                                                controlValue[prop.propertyName] = null;
                                                if (
                                                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                                                !control.controls[prop.propertyName]) {
                                                    // eslint-disable-next-line @typescript-eslint/no-explicit-any
                                                    control.addControl(prop.propertyName, formBuilder.control(controlValue[prop.propertyName]));
                                                }
                                            }
                                        }
                                    }
                                });
                                setValuesForControls(formBuilder, 
                                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                                control, controlValue);
                            }
                            setValidatorsToControls(formBuilder, control, rootFormGroup);
                            if (!rootFormGroup) {
                                throw Error('rootFormGroup not set');
                            }
                            if (!control.hasAsyncValidator(rootFormGroup.commonAsyncValidator)) {
                                control.setAsyncValidators(rootFormGroup.commonAsyncValidator);
                                /* control.updateValueAndValidity(); */
                            }
                        });
                        if (!rootFormGroup) {
                            throw Error('rootFormGroup not set');
                        }
                        if (!form.controls[metadataItem.propertyName].hasAsyncValidator(rootFormGroup.commonAsyncValidator)) {
                            form.controls[metadataItem.propertyName].setAsyncValidators(rootFormGroup.commonAsyncValidator);
                            /* (
                              form.controls[metadataItem.propertyName] as DynamicFormArray
                            ).updateValueAndValidity(); */
                        }
                    }
                    catch (err) {
                        console.log({ err, metadataItem, form });
                    }
                }
            }
        }
    });
}
function validateAllFormFields(form) {
    let control;
    let formArrayControlsLength;
    Object.keys(form.controls).forEach((field) => {
        control = form.get(field);
        // Control
        if (control instanceof FormControl) {
            control.markAsTouched({ onlySelf: true });
        }
        // Group: recursive
        else if (control instanceof FormGroup) {
            validateAllFormFields(control);
        }
        // Array
        else if (control instanceof FormArray) {
            formArrayControlsLength = control.controls.length;
            for (let i = 0; i < formArrayControlsLength; i++) {
                // Control in Array
                if (control.controls[i] instanceof FormControl) {
                    control.controls[i].markAsTouched({
                        onlySelf: true,
                    });
                }
                // Group in Array: recursive
                if (control.controls[i] instanceof FormGroup) {
                    validateAllFormFields(control.controls[i]);
                }
            }
        }
    });
}
// custom: getMetadata args have been changed. Before optional parameters, these were added:
// dynamicFormBuilderOptions
// currentDepth
// propertyPath
function getMetadata(classType, dynamicFormBuilderOptions, currentDepth, currentPath, classTransformOptions, isRoot) {
    const classTransformerMetadataStorage = 
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    getGlobal()['classTransformerMetadataStorage'] || undefined;
    if (!classTransformerMetadataStorage) {
        throw new Error('classTransformerMetadataStorage not set in windows, please use the "class-transformer-global-storage" instead of "class-transformer"');
    }
    const classValidatorMetadataStorage = 
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    getGlobal()['classValidatorMetadataStorage'] || undefined;
    // (window as any)['classValidatorMetadataStorage'] || undefined;
    if (!classValidatorMetadataStorage) {
        throw new Error('classValidatorMetadataStorage not set in windows');
    }
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    let exposeProperties = [];
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    let excludeProperties = [];
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    let prevMultiTypes = [];
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    let multiTypes = null;
    // need for create all link for multi types
    if (isRoot) {
        while (!prevMultiTypes ||
            !multiTypes ||
            prevMultiTypes.length !== multiTypes.length) {
            prevMultiTypes = multiTypes;
            plainToClass(classType, {}, classTransformOptions);
            multiTypes = Array.from(classTransformerMetadataStorage._typeMetadatas)
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                .map((meta) => Array.from(meta[1])
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                .filter((sub) => {
                return sub[1].options.discriminator;
            })
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                .map((sub) => {
                return { ...sub[1], ...sub[1].options.discriminator };
            }))
                .filter((arr) => arr.length > 0)
                .map((arr) => arr.map((item) => ({
                classType: item.target,
                propertyName: item.propertyName,
                types: item.subTypes,
                property: item.property,
            })))
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                .reduce((all, cur) => [...all, ...cur], []);
            (multiTypes || []).forEach((multiType) => multiType.types.forEach((type) => plainToClass(multiType.classType, { [multiType.propertyName]: { [multiType.property]: type.name } }, classTransformOptions)));
            exposeProperties =
                classTransformerMetadataStorage.getExposedMetadatas(classType);
        }
        prevMultiTypes = [];
        multiTypes = null;
        // need for create all link for multi types
        while (!prevMultiTypes ||
            !multiTypes ||
            prevMultiTypes.length !== multiTypes.length) {
            prevMultiTypes = multiTypes;
            plainToClass(classType, {}, classTransformOptions);
            multiTypes = Array.from(classTransformerMetadataStorage._excludeMetadatas)
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                .map((meta) => Array.from(meta[1])
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                .filter((sub) => {
                return sub[1].options.discriminator;
            })
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                .map((sub) => ({ ...sub[1], ...sub[1].options.discriminator })))
                .filter((arr) => arr.length > 0)
                .map((arr) => arr.map((item) => ({
                classType: item.target,
                propertyName: item.propertyName,
                types: item.subTypes,
                property: item.property,
            })))
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                .reduce((all, cur) => [...all, ...cur], []);
            (multiTypes || []).forEach((multiType) => multiType.types.forEach((type) => plainToClass(multiType.classType, { [multiType.propertyName]: { [multiType.property]: type.name } }, classTransformOptions)));
            excludeProperties =
                classTransformerMetadataStorage.getExcludedMetadatas(classType);
        }
    }
    else {
        exposeProperties =
            classTransformerMetadataStorage.getExposedMetadatas(classType);
        excludeProperties =
            classTransformerMetadataStorage.getExcludedMetadatas(classType);
    }
    // need for support ancestors multiple types of form group
    const ancestorsProperties = [
        Array.from(classTransformerMetadataStorage._ancestorsMap)
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            .filter((item) => item[1][0] === classType)
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            .map((item) => classTransformerMetadataStorage.getExposedMetadatas(item[0]))
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            .reduce((all, cur) => [...all, ...cur], []),
    ];
    const properties = [
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        ...exposeProperties,
        ...excludeProperties,
        ...(ancestorsProperties.length > 0 ? ancestorsProperties[0] : []),
    ].filter(
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    (existsLimit, existsIndex, items) => {
        return ((classTransformOptions?.excludeGroups
            ? !classTransformOptions?.excludeGroups.find((excludeGroup) => existsLimit?.options?.groups?.includes(excludeGroup))
            : true) &&
            existsIndex ===
                items.findIndex(
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                (it) => it.propertyName === existsLimit.propertyName));
    });
    const depthLimitExceeded = currentDepth >= dynamicFormBuilderOptions.maxNestedModelDepth;
    properties.forEach((exposeMetadataItem, index) => {
        const propertyMetadata = classTransformerMetadataStorage.findTypeMetadata(exposeMetadataItem.target, exposeMetadataItem.propertyName);
        // custom: getMetadata args have been changed. Before optional parameters, these were added:
        // dynamicFormBuilderOptions
        // currentDepth
        // propertyPath
        // old implementation:
        /*
        if (propertyMetadata) {
          properties[index] = {
            ...getMetadata(
              propertyMetadata.typeFunction(),
              classTransformOptions,
              false
            ),
            isArray: Array === propertyMetadata.reflectedType,
            propertyName: exposeMetadataItem.propertyName,
          }
        };
        */
        if (propertyMetadata) {
            // custom: skip if depthLimitExceeded
            if (depthLimitExceeded) {
                console.warn('ngx-dynamic-form-builder has hit depth limit!', propertyMetadata);
                delete properties[index];
            }
            else {
                const propertyPath = [...currentPath, exposeMetadataItem.propertyName];
                const assignProperty = () => {
                    properties[index] = {
                        ...getMetadata(propertyMetadata.typeFunction(), dynamicFormBuilderOptions, currentDepth + 1, propertyPath, undefined, currentDepth === 0),
                        isArray: Array === propertyMetadata.reflectedType,
                        propertyName: exposeMetadataItem.propertyName,
                    };
                };
                // custom: this is new logic. Instead of assigning all, 
                // metadata will be filtered out if allowedNestedModels are defined and do not include this prop.
                if (!Array.isArray(dynamicFormBuilderOptions.allowedNestedModels)) {
                    assignProperty();
                }
                else {
                    if (dynamicFormBuilderOptions.allowedNestedModels.includes(propertyPath.join('.'))) {
                        assignProperty();
                    }
                    else {
                        delete properties[index];
                    }
                }
            }
        }
        else {
            properties[index] = {
                classType: null,
                properties: [],
                isArray: false,
                propertyName: exposeMetadataItem.propertyName,
            };
        }
    });
    const metadata = {
        classType,
        properties: properties.filter(prop => typeof prop !== 'undefined'),
        withAncestors: (ancestorsProperties?.length > 0 ? ancestorsProperties[0] : []).length >
            0,
        isArray: false,
        propertyName: null,
    };
    return metadata;
}
function setupClassTransformMetadata({ defaultMetadata, classType, dynamicForm, dynamicFormBuilderOptions, }) {
    if (!defaultMetadata && classType) {
        dynamicForm.classTransformMetadata = getMetadata(classType, dynamicForm.dynamicFormBuilderOptions, 0, [], dynamicFormBuilderOptions.classTransformOptions, true);
    }
    else {
        dynamicForm.classTransformMetadata =
            defaultMetadata;
    }
    if (!dynamicForm.classTransformMetadata) {
        throw new Error('metadata not set');
    }
}
function setupDynamicFormBuilderOptions({ dynamicForm, dynamicFormBuilderOptions, }) {
    if (!dynamicForm) {
        throw new Error('dynamicForm not set');
    }
    if (!dynamicForm.dynamicFormBuilderOptions) {
        dynamicForm.dynamicFormBuilderOptions = dynamicFormBuilderOptions || {};
    }
    dynamicForm.dynamicFormBuilderOptions.classValidatorOptions = {
        ...(dynamicForm.dynamicFormBuilderOptions.classValidatorOptions || {}),
    };
    dynamicForm.dynamicFormBuilderOptions.classTransformOptions = {
        ...DEFAULT_CLASS_TRANSFORM_OPTIONS,
        ...(dynamicForm.dynamicFormBuilderOptions.classTransformOptions || {}),
    };
    dynamicForm.dynamicFormBuilderOptions.classTransformToPlainOptions = {
        ...(dynamicForm.dynamicFormBuilderOptions.classTransformToPlainOptions ||
            {}),
    };
    dynamicForm.dynamicFormBuilderOptions.maxNestedModelDepth =
        (typeof dynamicForm.dynamicFormBuilderOptions.maxNestedModelDepth === 'number' ?
            dynamicForm.dynamicFormBuilderOptions.maxNestedModelDepth : 5);
    // dynamicFormBuilderOptions.allowedNestedModels defaults to undefined
}
function createAllFormGroupChildrenControls({ dynamicForm, defaultValue, formBuilder, rootFormGroup, dynamicFormBuilderOptions, }) {
    dynamicForm.classTransformMetadata.properties.forEach((metadataItem, index) => {
        dynamicForm.classTransformMetadata.properties[index].ignore = true;
        if (metadataItem.propertyName) {
            if (Object.keys(defaultValue || {}).length === 0 ||
                Object.getOwnPropertyDescriptor(defaultValue || {}, metadataItem.propertyName) ||
                metadataItem.isArray) {
                if (!metadataItem.isArray) {
                    if (!metadataItem.classType ||
                        isPrimitiveClass(metadataItem.classType)) {
                        addDynamicPrimitiveObjectAsFormGroup({
                            defaultValue,
                            metadataItem,
                            formBuilder,
                            dynamicForm,
                            index,
                        });
                    }
                    else {
                        addDynamicTypedObjectAsFormGroup({
                            defaultValue,
                            metadataItem,
                            formBuilder,
                            rootFormGroup,
                            dynamicFormBuilderOptions,
                            dynamicForm,
                            index,
                        });
                    }
                }
                else {
                    if (!metadataItem.classType ||
                        isPrimitiveClass(metadataItem.classType)) {
                        addDynamicPrimitiveObjectAsFormArray({
                            defaultValue,
                            metadataItem,
                            formBuilder,
                            dynamicForm,
                            index,
                        });
                    }
                    else {
                        addDynamicTypedObjectAsFormArray({
                            defaultValue,
                            metadataItem,
                            formBuilder,
                            rootFormGroup,
                            dynamicFormBuilderOptions,
                            dynamicForm,
                            index,
                        });
                    }
                }
            }
        }
    });
}
function addDynamicTypedObjectAsFormArray({ defaultValue, metadataItem, formBuilder, rootFormGroup, dynamicFormBuilderOptions, dynamicForm, index, }) {
    if (!metadataItem.propertyName) {
        throw new Error('propertyName not set');
    }
    const defaultArray = defaultValue[metadataItem.propertyName];
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const value = defaultArray && defaultArray.length > 0 ? defaultArray : [];
    const control = formBuilder.array([]);
    value.forEach((item) => {
        const form = formBuilder.group(item);
        form.setParent(control);
        control.push(createFormControls({
            classType: metadataItem.classType,
            formBuilder,
            form,
            rootFormGroup,
            metadata: metadataItem,
            defaultValue: item,
            dynamicFormBuilderOptions,
        }));
    });
    dynamicForm.addControl(metadataItem.propertyName, control);
    control.classTransformMetadata = metadataItem;
    control.classTransformMetadata.parent = dynamicForm.classTransformMetadata;
    dynamicForm.classTransformMetadata.properties[index].ignore = false;
}
function addDynamicPrimitiveObjectAsFormArray({ defaultValue, metadataItem, formBuilder, dynamicForm, index, }) {
    if (!metadataItem.propertyName) {
        throw new Error('propertyName not set');
    }
    const defaultArray = defaultValue[metadataItem.propertyName];
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const value = defaultArray && defaultArray.length > 0 ? defaultArray : [];
    const control = formBuilder.array(value);
    // todo: incorrect work with addControl
    dynamicForm.setControl(metadataItem.propertyName, control);
    control.classTransformMetadata = metadataItem;
    control.classTransformMetadata.parent = dynamicForm.classTransformMetadata;
    dynamicForm.classTransformMetadata.properties[index].ignore = false;
}
function addDynamicTypedObjectAsFormGroup({ defaultValue, metadataItem, formBuilder, rootFormGroup, dynamicFormBuilderOptions, dynamicForm, index, }) {
    if (!metadataItem.propertyName) {
        throw new Error('propertyName not set');
    }
    const value = defaultValue[metadataItem.propertyName] ||
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        metadataItem.properties.reduce((all, cur) => {
            if (cur.propertyName) {
                if (cur.isArray) {
                    all[cur.propertyName] = [];
                }
                else {
                    if (!cur.classType || isPrimitiveClass(cur.classType)) {
                        all[cur.propertyName] = null;
                    }
                    else {
                        all[cur.propertyName] = {};
                    }
                }
            }
            return all;
        }, {});
    const form = formBuilder.group({});
    form.setParent(rootFormGroup);
    const control = createFormControls({
        classType: metadataItem.classType,
        formBuilder,
        form,
        rootFormGroup,
        metadata: metadataItem,
        defaultValue: value,
        dynamicFormBuilderOptions,
    });
    dynamicForm.addControl(metadataItem.propertyName, control);
    control.classTransformMetadata = metadataItem;
    control.classTransformMetadata.parent = dynamicForm.classTransformMetadata;
    dynamicForm.classTransformMetadata.properties[index].ignore = false;
}
function addDynamicPrimitiveObjectAsFormGroup({ defaultValue, metadataItem, formBuilder, dynamicForm, index, }) {
    if (!metadataItem.propertyName) {
        throw new Error('propertyName not set');
    }
    const value = Object.getOwnPropertyDescriptor(defaultValue, metadataItem.propertyName)
        ? defaultValue[metadataItem.propertyName]
        : undefined;
    const control = formBuilder.control(value);
    dynamicForm.addControl(metadataItem.propertyName, control);
    control.classTransformMetadata = metadataItem;
    control.classTransformMetadata.parent = dynamicForm.classTransformMetadata;
    dynamicForm.classTransformMetadata.properties[index].ignore = false;
}
function addDynamicMethodsToRootFormGroup({ dynamicForm, rootFormGroup, formBuilder, }) {
    dynamicForm.externalErrorsSubject = new BehaviorSubject({});
    dynamicForm.patchDynamicFormBuilderOptions = (dynamicFormBuilderOptions) => {
        if (!dynamicForm) {
            throw new Error('rootFormGroup not set');
        }
        dynamicForm.dynamicFormBuilderOptions = {
            ...dynamicForm.dynamicFormBuilderOptions,
            ...dynamicFormBuilderOptions,
        };
        setValidatorsToControls(formBuilder, dynamicForm);
        dynamicForm.patchValue(dynamicForm.value, { emitEvent: false });
    };
    dynamicForm.getObject = () => {
        if (!dynamicForm.classTransformMetadata.classType) {
            throw new Error('classType not set');
        }
        return plainToClass(dynamicForm.classTransformMetadata.classType, {
            ...dynamicForm.value,
        }, dynamicForm.dynamicFormBuilderOptions.classTransformOptions);
    };
    dynamicForm.setObject = (object) => {
        setValuesForControls(formBuilder, 
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        dynamicForm, classToPlain(object, dynamicForm.dynamicFormBuilderOptions.classTransformToPlainOptions));
        setValidatorsToControls(formBuilder, dynamicForm, rootFormGroup);
    };
    dynamicForm.refresh = () => {
        setValidatorsToControls(formBuilder, dynamicForm, rootFormGroup);
        dynamicForm.patchValue(dynamicForm.value);
    };
    dynamicForm.getJSON = () => {
        return recursiveRemoveDynamicControlOptions(cloneDeep(dynamicForm.value));
    };
    dynamicForm.setJSON = (value, options) => {
        setValuesForControls(formBuilder, dynamicForm, value);
        setValidatorsToControls(formBuilder, dynamicForm, rootFormGroup);
        dynamicForm.patchValue(dynamicForm.value, options);
    };
    dynamicForm.setExternalErrors = (externalErrors) => {
        dynamicForm.externalErrorsSubject.next(recursiveRemoveDynamicControlOptions(cloneDeep(externalErrors)));
        dynamicForm.patchValue(dynamicForm.value, { emitEvent: false });
    };
    dynamicForm.getExternalErrors = () => {
        return recursiveRemoveDynamicControlOptions(cloneDeep(dynamicForm.externalErrorsSubject.value));
    };
    const dynamicFormProxy = new Proxy(dynamicForm, {
        get(target, prop) {
            if (prop === 'object') {
                return target.getObject();
            }
            if (prop === 'json') {
                return target.getJSON();
            }
            if (prop === 'externalErrors') {
                return target.getExternalErrors();
            }
            return target[prop];
        },
        set(target, prop, 
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        value) {
            if (prop === 'object') {
                target.setObject(value);
                return true;
            }
            if (prop === 'json') {
                target.setJSON(value);
                return true;
            }
            if (prop === 'externalErrors') {
                target.setExternalErrors(value);
                return true;
            }
            Object.defineProperty(target, prop, {
                value,
                configurable: true,
            });
            return true;
        },
    });
    dynamicFormProxy.patchValue(dynamicFormProxy.value, { emitEvent: false });
    setValidatorsToControls(formBuilder, dynamicFormProxy, rootFormGroup);
    return dynamicFormProxy;
}
function addDynamicMethodsToFormGroup(dynamicForm) {
    dynamicForm.getClassValidatorErrors = (control, nullIfEmpty) => {
        const dynamicControl = control;
        if (!dynamicControl?.classValidatorErrorsSubject) {
            return of(null);
        }
        return dynamicControl.classValidatorErrorsSubject.pipe(map((errors) => {
            if (dynamicControl?.classTransformMetadata?.propertyName &&
                Object.getOwnPropertyDescriptor(errors, dynamicControl?.classTransformMetadata?.propertyName)) {
                return errors[dynamicControl?.classTransformMetadata?.propertyName]
                    ?.messages?.length === 0 && nullIfEmpty
                    ? null
                    : errors[dynamicControl?.classTransformMetadata?.propertyName];
            }
            return null;
        }));
    };
}
function addCommonAsyncValidatorToRootForm(rootFormGroup) {
    if (!rootFormGroup.commonAsyncValidator) {
        rootFormGroup.commonAsyncValidatorSubject = new ReplaySubject();
        rootFormGroup.commonAsyncValidator = (control) => {
            const metadata = control
                ?.classTransformMetadata;
            if (!metadata) {
                throw new Error('classTransformMetadata not found in form group');
            }
            const rootValue = rootFormGroup ? rootFormGroup.value : null;
            const { inputPath, errorsPath, controlPath } = findControlPaths(control);
            const object = setDynamicControlValue(rootFormGroup, rootValue, controlPath, control, metadata);
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            setDynamicControlOptions(object, {
                rootValue,
                controlPath,
                control,
                metadata,
            });
            const localClassValidatorOptions = cloneDeep(rootFormGroup.dynamicFormBuilderOptions.classValidatorOptions);
            const globalClassValidatorOptions = cloneDeep(getGlobalDynamicFormBuilderOptions().classValidatorOptions);
            const classValidatorOptions = mergeWith(globalClassValidatorOptions, localClassValidatorOptions);
            return (rootFormGroup.commonAsyncValidatorFirstChanged
                ? from(validate(object, classValidatorOptions))
                : of(validateSync(object, classValidatorOptions))).pipe(debounceTime(700), map((rootClassValidatorErrors) => {
                setRootFormGroupClassValidatorErrors(rootFormGroup, rootClassValidatorErrors);
                const dynamicFormControl = setFormControlClassValidatorErrors(control, rootFormGroup, errorsPath, inputPath);
                return Object.keys(dynamicFormControl.classValidatorErrors || {})
                    .length > 0
                    ? dynamicFormControl.classValidatorErrors
                    : null;
            }));
        };
    }
}
function setDynamicControlValue(rootFormGroup, 
// eslint-disable-next-line @typescript-eslint/no-explicit-any
rootValue, controlPath, control, metadata) {
    if (!rootFormGroup) {
        throw Error('rootFormGroup not set');
    }
    if (!rootFormGroup.classTransformMetadata.classType) {
        throw Error('classType not set');
    }
    // need for resize arrays
    lodashSet(rootValue, controlPath, control.value);
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const object = plainToClass(rootFormGroup.classTransformMetadata.classType, rootValue, rootFormGroup.dynamicFormBuilderOptions.classTransformOptions);
    // set original values of inputs
    if (metadata.isArray &&
        (!metadata.classType || isPrimitiveClass(metadata.classType))) {
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        control.value.forEach((item, index) => {
            lodashSet(object, `${controlPath}.${index}`, control.value[index]);
        });
    }
    // need for correct work ValidateIfUpdater
    if (metadata.parent?.classType && metadata.propertyIndex === undefined) {
        replaceLastPropertyToDynamicControlOptionsByDotPath(controlPath, object, rootValue, control, metadata);
    }
    return object;
}
function findControlPaths(control) {
    const angularFormValidatorPath = [];
    const inputPath = [];
    const errorsPath = [];
    let ctrl = control;
    while (ctrl) {
        const key = ctrl.classTransformMetadata?.propertyIndex?.toString() ||
            ctrl.classTransformMetadata?.propertyName ||
            null;
        const parent = ctrl.parent;
        if (key !== null) {
            inputPath.push(key);
            errorsPath.push(key);
            if (parent?.parent) {
                errorsPath.push('children');
            }
            if (!parent?.classTransformMetadata?.isArray) {
                angularFormValidatorPath.push(key);
            }
        }
        ctrl = parent;
    }
    const controlPath = inputPath.reverse().join('.');
    return {
        inputPath,
        errorsPath,
        controlPath,
        angularFormValidatorPath: angularFormValidatorPath.reverse().join('.'),
    };
}
function setFormControlClassValidatorErrors(control, rootFormGroup, errorsPath, inputPath) {
    const dynamicFormControl = control;
    const propErrors = lodashGet(rootFormGroup.classValidatorErrors, errorsPath.reverse());
    dynamicFormControl.classValidatorErrors =
        Object.keys(propErrors || {}).length > 0
            ? {
                [inputPath.reverse()[0]]: propErrors || {},
            }
            : {};
    if (!dynamicFormControl.classValidatorErrorsSubject) {
        dynamicFormControl.classValidatorErrors = {};
        dynamicFormControl.classValidatorErrorsSubject =
            new BehaviorSubject({});
    }
    dynamicFormControl.classValidatorErrorsSubject.next(dynamicFormControl.classValidatorErrors);
    if (!dynamicFormControl.customValidateErrors) {
        dynamicFormControl.customValidateErrors =
            new BehaviorSubject({});
    }
    dynamicFormControl.customValidateErrors.next(transformClassValidatorErrorsToShortValidationErrors(dynamicFormControl.classValidatorErrors));
    return dynamicFormControl;
}
function setRootFormGroupClassValidatorErrors(rootFormGroup, rootClassValidatorErrors) {
    if (!rootFormGroup) {
        throw new Error('rootFormGroup not set');
    }
    rootFormGroup.classValidatorErrors =
        mergeErrors(rootFormGroup.externalErrorsSubject?.value || {}, transformValidationErrorsToClassValidatorErrors(rootClassValidatorErrors) || {}) || {};
    if (!rootFormGroup.classValidatorErrorsSubject) {
        rootFormGroup.classValidatorErrors = {};
        rootFormGroup.classValidatorErrorsSubject =
            new BehaviorSubject({});
    }
    rootFormGroup.classValidatorErrorsSubject.next(rootFormGroup.classValidatorErrors);
    if (!rootFormGroup.customValidateErrors) {
        rootFormGroup.customValidateErrors =
            new BehaviorSubject({});
    }
    rootFormGroup.customValidateErrors.next(mergeErrors(collectDynamicFormGroupErrors(rootFormGroup), transformClassValidatorErrorsToShortValidationErrors(rootFormGroup.classValidatorErrors)));
}
function subscribeToRootFormGroupValueChanges(rootFormGroup) {
    if (!rootFormGroup.commonAsyncValidatorFirstChangedSubscription) {
        rootFormGroup.commonAsyncValidatorFirstChangedSubscription =
            rootFormGroup?.valueChanges
                .pipe(tap(() => {
                if (!rootFormGroup) {
                    throw Error('rootFormGroup not set');
                }
                rootFormGroup.commonAsyncValidatorFirstChanged = true;
            }), delay(100), tap(() => {
                if (!rootFormGroup) {
                    throw Error('rootFormGroup not set');
                }
                rootFormGroup.customValidateErrors.next(mergeErrors(collectDynamicFormGroupErrors(rootFormGroup), transformClassValidatorErrorsToShortValidationErrors(rootFormGroup.classValidatorErrors)));
            }), catchError((err) => {
                console.log({ err, rootFormGroup });
                return of(null);
            }))
                .subscribe();
    }
}
function subscribeToGlobalDynamicFormBuilderOptionsChange(formBuilder, rootFormGroup) {
    if (!rootFormGroup.globalDynamicFormBuilderOptionsChangedSubscription) {
        rootFormGroup.globalDynamicFormBuilderOptionsChangedSubscription =
            getGlobalDynamicFormBuilderOptionsSubject()
                .pipe(tap(() => {
                setValidatorsToControls(formBuilder, rootFormGroup);
                rootFormGroup.patchValue(rootFormGroup.value, { emitEvent: false });
            }), catchError((err) => {
                console.log({ err, rootFormGroup });
                return of(null);
            }))
                .subscribe();
    }
}

var DynamicFormBuilderOptionsKeys;
(function (DynamicFormBuilderOptionsKeys) {
    DynamicFormBuilderOptionsKeys["classValidatorOptions"] = "classValidatorOptions";
    DynamicFormBuilderOptionsKeys["classTransformOptions"] = "classTransformOptions";
    DynamicFormBuilderOptionsKeys["classTransformToPlainOptions"] = "classTransformToPlainOptions";
    DynamicFormBuilderOptionsKeys["angularValidators"] = "angularValidators";
    DynamicFormBuilderOptionsKeys["maxNestedModelDepth"] = "maxNestedModelDepth";
    DynamicFormBuilderOptionsKeys["allowedNestedModels"] = "allowedNestedModels";
})(DynamicFormBuilderOptionsKeys || (DynamicFormBuilderOptionsKeys = {}));

class DynamicFormBuilder extends FormBuilder {
    rootFormGroup(classType, 
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    controlsConfig, 
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    options) {
        let defaultValue = controlsConfig || {};
        let dynamicFormBuilderOptions = options;
        if (!options &&
            Object.keys(controlsConfig || {}).find((controlName) => Object.keys(DynamicFormBuilderOptionsKeys).find((optionsKey) => optionsKey === controlName))) {
            dynamicFormBuilderOptions = controlsConfig;
            defaultValue = {};
        }
        const form = createFormControls({
            classType,
            form: super.group({}, dynamicFormBuilderOptions?.angularValidators ||
                dynamicFormBuilderOptions),
            formBuilder: this,
            defaultValue,
            dynamicFormBuilderOptions,
        });
        return form;
    }
    childFormGroup(
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    rootFormGroup, classType, 
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    controlsConfig) {
        const defaultValue = controlsConfig || {};
        const form = createFormControls({
            classType,
            form: super.group({}),
            formBuilder: this,
            defaultValue,
            rootFormGroup,
        });
        return form;
    }
}

function ValidateIfUpdater(updaterFn) {
    return Validate(dynamicValidateIfUpdater(updaterFn));
}
function dynamicValidateIfUpdater(updaterFn) {
    let ValidateIfUpdaterConstraint = class ValidateIfUpdaterConstraint {
        validate(value, validationArguments) {
            const dynamicFormBuilderOptions = getDynamicControlOptions(validationArguments?.object);
            const rootFormGroup = dynamicFormBuilderOptions?.control
                .root;
            if (rootFormGroup) {
                const key = `${ValidateIfUpdater.name}_${dynamicFormBuilderOptions?.controlPath}`;
                const prevValue = getCustomDataToRootFormGroup(rootFormGroup, key, null);
                if (value !== prevValue) {
                    setTimeout(() => {
                        const appiled = updaterFn(dynamicFormBuilderOptions?.controlPath, value, rootFormGroup, dynamicFormBuilderOptions);
                        if (appiled) {
                            setCustomDataToRootFormGroup(rootFormGroup, key, value);
                        }
                    }, 100);
                }
            }
            return true;
        }
    };
    ValidateIfUpdaterConstraint = __decorate([
        ValidatorConstraint()
    ], ValidateIfUpdaterConstraint);
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    return ValidateIfUpdaterConstraint;
}

// for correct work expose with js object
// https://github.com/typestack/class-transformer/issues/365
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function ExposeNested(options) {
    const exposeFn = Expose(options);
    const transformFn = (propertyKey) => Transform((params) => {
        return ((params.obj && Object.getOwnPropertyDescriptor(params.obj, propertyKey)
            ? params.obj[propertyKey]
            : options?.default) || params.value);
    });
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    return (target, propertyKey) => {
        transformFn(propertyKey)(target, propertyKey);
        exposeFn(target, propertyKey);
    };
}

/**
 * Generated bundle index. Do not edit.
 */

export { DEFAULT_CLASS_TRANSFORM_OPTIONS, DynamicFormBuilder, DynamicFormBuilderOptionsKeys, ExposeNested, ValidateIfUpdater, __dynamicControlOptions__, collectDynamicFormGroupErrors, createFormControls, dynamicValidateIfUpdater, getCustomDataToRootFormGroup, getDynamicControlOptions, getGlobal, getGlobalDynamicFormBuilderOptions, getGlobalDynamicFormBuilderOptionsSubject, isPrimitiveClass, isPrimitiveType, mergeErrors, recursiveRemoveDynamicControlOptions, removeDynamicControlOptions, replaceLastPropertyToDynamicControlOptionsByDotPath, setCustomDataToRootFormGroup, setDynamicControlOptions, setGlobalDynamicFormBuilderOptions, setValidatorsToControls, setValuesForControls, transformClassValidatorErrorsToShortValidationErrors, transformValidationErrorsToClassValidatorErrors, validateAllFormFields };
//# sourceMappingURL=yoolabs-ngx-dynamic-form-builder.mjs.map
